from __future__ import annotations

from collections.abc import Mapping
from typing import Any

__all__ = ("plugin",)

from pickle import PickleBuffer

import awkward as ak
from awkward.typetracer import PlaceholderArray


def _maybe_make_pickle_buffer(buffer: Any) -> PlaceholderArray | PickleBuffer:
    if isinstance(buffer, PlaceholderArray):
        return buffer
    else:
        return PickleBuffer(buffer)


def _without_transient_attrs(attrs: Mapping[str, Any]) -> Mapping[str, Any]:
    return {k: v for k, v in attrs.items() if not k.startswith("@")}


def _unpickle_record_schema_1(
    form_dict: dict,
    length: Any,
    container: Mapping[str, Any],
    behavior: Mapping | None,
    attrs: Mapping[str, Any] | None,
    at: int,
) -> ak.Record:
    array_layout = ak.from_buffers(
        form_dict, length, container, behavior=behavior, attrs=attrs, highlevel=False
    )
    layout = ak.record.Record(array_layout, at)
    return ak.Record(layout, behavior=behavior, attrs=attrs)


def _unpickle_array_schema_1(
    form_dict: dict,
    length: Any,
    container: Mapping[str, Any],
    behavior: Mapping | None,
    attrs: Mapping[str, Any] | None,
) -> ak.Array:
    return ak.from_buffers(
        form_dict, length, container, behavior=behavior, attrs=attrs, highlevel=True
    )


def pickle_record(record: ak.Record, protocol: int) -> tuple:
    layout = ak.to_layout(record, allow_record=True)
    form, length, container = ak.operations.to_buffers(
        layout.array,
        buffer_key="{form_key}-{attribute}",
        form_key="node{id}",
        byteorder="<",
    )

    # For pickle >= 5, we can avoid copying the buffers
    if protocol >= 5:
        container = {k: _maybe_make_pickle_buffer(v) for k, v in container.items()}

    if record.behavior is ak.behavior:
        behavior = None
    else:
        behavior = record.behavior

    if record._attrs is None:
        attrs = record._attrs
    else:
        attrs = _without_transient_attrs(record._attrs)

    return (
        _unpickle_record_schema_1,
        (form.to_dict(), length, container, behavior, attrs, layout.at),
    )


def pickle_array(array: ak.Array, protocol: int) -> tuple:
    layout = ak.to_layout(array, allow_record=False)
    form, length, container = ak.operations.to_buffers(
        layout,
        buffer_key="{form_key}-{attribute}",
        form_key="node{id}",
        byteorder="<",
    )

    # For pickle >= 5, we can avoid copying the buffers
    if protocol >= 5:
        container = {k: _maybe_make_pickle_buffer(v) for k, v in container.items()}

    if array.behavior is ak.behavior:
        behavior = None
    else:
        behavior = array.behavior

        if array._attrs is None:
            attrs = array._attrs
        else:
            attrs = _without_transient_attrs(array._attrs)

    return (
        _unpickle_array_schema_1,
        (form.to_dict(), length, container, behavior, attrs),
    )


def plugin(obj: Any, protocol: int) -> tuple:
    if isinstance(obj, ak.Record):
        return pickle_record(obj, protocol)
    elif isinstance(obj, ak.Array):
        return pickle_array(obj, protocol)
    else:
        return NotImplemented



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

import awkward as ak
from awkward.operations.ak_concatenate import (
    enforce_concatenated_form as enforce_layout_to_concatenated_form,
)
from awkward.typetracer import typetracer_from_form
from dask.base import tokenize
from dask.highlevelgraph import HighLevelGraph

from dask_awkward.layers import AwkwardMaterializedLayer
from dask_awkward.lib.core import (
    Array,
    PartitionCompatibility,
    map_partitions,
    new_array_object,
    partition_compatibility,
)
from dask_awkward.utils import DaskAwkwardNotImplemented, IncompatiblePartitions

if TYPE_CHECKING:
    from awkward.forms import Form
    from awkward.highlevel import Array as AwkwardArray


class ConcatenateFnAxisGT0:
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, *args):
        return ak.concatenate(list(args), **self.kwargs)


def _enforce_concatenated_form(array: AwkwardArray, form: Form) -> AwkwardArray:
    layout = ak.to_layout(array)
    # TODO: should this check whether the form agrees first, or assume that the
    #       operation is harmless if not required?
    result = enforce_layout_to_concatenated_form(layout, form)
    return ak.Array(result, behavior=array._behavior, attrs=array._attrs)


def _concatenate_axis_0_meta(*arrays: AwkwardArray) -> AwkwardArray:
    # At this stage, the metas have all been enforced to the same type
    return arrays[0]


def concatenate(
    arrays: list[Array],
    axis: int = 0,
    mergebool: bool = True,
    highlevel: bool = True,
    behavior: Mapping | None = None,
    attrs: Mapping[str, Any] | None = None,
) -> Array:
    label = "concatenate"
    token = tokenize(arrays, axis, mergebool, highlevel, behavior)
    name = f"{label}-{token}"

    metas = [c._meta for c in arrays]

    if len(metas) == 0:
        raise ValueError("Need at least one array to concatenate")

    # Are we performing a _logical_ concatenation?
    if axis == 0:
        # There are two possible cases here:
        # 1. all arrays have identical metas — just grow the Dask collection
        # 2. some arrays have different metas — coerce arrays to same form

        # Drop reports from metas to avoid later touching any buffers
        metas_no_report = [
            typetracer_from_form(x.layout.form, behavior=x.behavior, attrs=x._attrs)
            for x in metas
        ]
        # Concatenate metas to determine result form
        meta_no_report = ak.concatenate(
            metas_no_report, axis=0, behavior=behavior, attrs=attrs
        )
        intended_form = meta_no_report.layout.form

        # If any forms aren't equal to this form, we must enforce each form to the same type
        if any(
            not m.layout.form.is_equal_to(
                intended_form, all_parameters=True, form_key=False
            )
            for m in metas
        ):
            arrays = [
                map_partitions(
                    _enforce_concatenated_form,
                    c,
                    label="enforce-concat-form",
                    form=intended_form,
                    output_divisions=1,
                )
                for c in arrays
            ]

            g = {
                (name, i): k
                for i, k in enumerate(
                    k for collection in arrays for k in collection.__dask_keys__()
                )
            }

            aml = AwkwardMaterializedLayer(
                g,
                previous_layer_names=[a.name for a in arrays],
                fn=_concatenate_axis_0_meta,
            )
        else:
            g = {
                (name, i): k
                for i, k in enumerate(
                    k for collection in arrays for k in collection.__dask_keys__()
                )
            }

            aml = AwkwardMaterializedLayer(g, previous_layer_names=[arrays[0].name])

        hlg = HighLevelGraph.from_collections(name, aml, dependencies=arrays)
        return new_array_object(
            hlg,
            name,
            meta=meta_no_report,
            npartitions=sum(a.npartitions for a in arrays),
        )

    if axis > 0:
        if partition_compatibility(*arrays) == PartitionCompatibility.NO:
            raise IncompatiblePartitions("concatenate", *arrays)

        fn = ConcatenateFnAxisGT0(axis=axis, behavior=behavior, attrs=attrs)
        return map_partitions(fn, *arrays, label="concatenate-axisgt0")

    raise DaskAwkwardNotImplemented("TODO")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, NamedTuple

import numpy as np
from dask.base import unpack_collections

from dask_awkward.layers import AwkwardInputLayer

if TYPE_CHECKING:
    from awkward.highlevel import Array as AwkArray

    from dask_awkward.lib.core import Array


class NecessaryBuffers(NamedTuple):
    data_and_shape: frozenset[str]
    shape_only: frozenset[str]


def report_necessary_buffers(
    *args: Any, traverse: bool = True
) -> dict[str, NecessaryBuffers | None]:
    r"""Determine the buffer keys necessary to compute a collection.

    Parameters
    ----------
    *args : Dask collections or HighLevelGraphs
        The collection (or collection graph) of interest. These can be
        individual objects, lists, sets, or dictionaries.
    traverse : bool, optional
        If True (default), builtin Python collections are traversed
        looking for any Dask collections they might contain.

    Returns
    -------
    dict[str, NecessaryBuffers | None]
        Mapping that pairs the input layers in the graph to objects
        describing the data and shape buffers that have been tagged
        as required by column optimisation of the given layer.

    Examples
    --------
    If we have a hypothetical parquet dataset (``ds``) with the fields

    - "foo"
    - "bar"
    - "baz"

    And the "baz" field has fields

    - "x"
    - "y"

    The calculation of ``ds.bar + ds.baz.x`` will only require the
    ``bar`` and ``baz.x`` columns from the parquet file.

    >>> import dask_awkward as dak
    >>> ds = dak.from_parquet("some-dataset")
    >>> ds.fields
    ["foo", "bar", "baz"]
    >>> ds.baz.fields
    ["x", "y"]
    >>> x = ds.bar + ds.baz.x
    >>> dak.report_necessary_buffers(x)
    {
        "from-parquet-abc123": NecessaryBuffers(
            data_and_shape=frozenset(...), shape_only=frozenset(...)
        )
    }

    """
    import dask_awkward.lib.optimize as o

    collections, _ = unpack_collections(*args, traverse=traverse)
    if not collections:
        return {}

    seen_names = set()

    name_to_necessary_buffers: dict[str, NecessaryBuffers | None] = {}
    for obj in collections:
        dsk = obj.__dask_graph__()
        keys = obj.__dask_keys__()
        projection_data = o._prepare_buffer_projection(dsk, keys)

        # If the projection failed, or there are no input layers
        if projection_data is None:
            # Ensure that we have a record of the seen layers, if they're inputs
            for name, layer in dsk.items():
                if isinstance(layer, AwkwardInputLayer):
                    seen_names.add(name)
            continue

        # Unpack projection information
        layer_to_reports, _ = projection_data
        for name, report in layer_to_reports.items():
            existing_buffers = name_to_necessary_buffers.setdefault(
                name, NecessaryBuffers(frozenset(), frozenset())
            )
            # Compute the shape-only keys in addition to the data and shape
            data_and_shape = frozenset(report.data_touched)
            shape_only = frozenset(report.shape_touched) - data_and_shape

            # Update set of touched keys
            assert existing_buffers is not None
            name_to_necessary_buffers[name] = NecessaryBuffers(
                data_and_shape=existing_buffers.data_and_shape | data_and_shape,
                shape_only=existing_buffers.shape_only | shape_only,
            )

    # Populate result with names of seen layers
    for k in seen_names:
        name_to_necessary_buffers.setdefault(k, None)
    return name_to_necessary_buffers


def report_necessary_columns(
    *args: Any, traverse: bool = True
) -> dict[str, frozenset[str] | None]:
    r"""Get columns necessary to compute a collection

    This function is specific to sources that are columnar (e.g. Parquet).

    Parameters
    ----------
    *args : Dask collections or HighLevelGraphs
        The collection (or collection graph) of interest. These can be
        individual objects, lists, sets, or dictionaries.
    traverse : bool, optional
        If True (default), builtin Python collections are traversed
        looking for any Dask collections they might contain.

    Returns
    -------
    dict[str, frozenset[str] | None]
        Mapping that pairs the input layers in the graph to the
        set of necessary IO columns that have been identified by column
        optimisation of the given layer. If the layer is not backed by a
        columnar source, then None is returned instead of a set.

    Examples
    --------
    If we have a hypothetical parquet dataset (``ds``) with the fields

    - "foo"
    - "bar"
    - "baz"

    And the "baz" field has fields

    - "x"
    - "y"

    The calculation of ``ds.bar + ds.baz.x`` will only require the
    ``bar`` and ``baz.x`` columns from the parquet file.

    >>> import dask_awkward as dak
    >>> ds = dak.from_parquet("some-dataset")
    >>> ds.fields
    ["foo", "bar", "baz"]
    >>> ds.baz.fields
    ["x", "y"]
    >>> x = ds.bar + ds.baz.x
    >>> dak.report_necessary_columns(x)
    {
        "from-parquet-abc123": frozenset({"bar", "baz.x"})
    }

    """
    import dask_awkward.lib.optimize as o

    collections, _ = unpack_collections(*args, traverse=traverse)
    if not collections:
        return {}

    seen_names = set()

    name_to_necessary_columns: dict[str, frozenset | None] = {}
    for obj in collections:
        dsk = obj.__dask_graph__()
        keys = obj.__dask_keys__()
        projection_data = o._prepare_buffer_projection(dsk, keys)

        # If the projection failed, or there are no input layers
        if projection_data is None:
            # Ensure that we have a record of the seen layers, if they're inputs
            for name, layer in dsk.items():
                if isinstance(layer, AwkwardInputLayer):
                    seen_names.add(name)
            continue

        # Unpack projection information
        layer_to_reports, layer_to_projection_state = projection_data
        for name, report in layer_to_reports.items():
            layer = dsk.layers[name]
            if not (isinstance(layer, AwkwardInputLayer) and layer.is_columnar):
                continue

            existing_columns = name_to_necessary_columns.setdefault(name, frozenset())

            assert existing_columns is not None
            # Update set of touched keys
            name_to_necessary_columns[name] = (
                existing_columns
                | layer.necessary_columns(
                    report=report, state=layer_to_projection_state[name]
                )
            )

    # Populate result with names of seen layers
    for k in seen_names:
        name_to_necessary_columns.setdefault(k, None)
    return name_to_necessary_columns


def _random_boolean_like(array_like: AwkArray, probability: float) -> AwkArray:
    import awkward as ak

    backend = ak.backend(array_like)
    layout = ak.to_layout(array_like)

    if ak.backend(array_like) == "typetracer":
        return ak.Array(
            ak.to_layout(np.empty(0, dtype=np.bool_)).to_typetracer(forget_length=True),
            behavior=array_like.behavior,
        )
    else:
        return ak.Array(
            np.random.random(layout.length) < probability,
            behavior=array_like.behavior,
            backend=backend,
        )


def sample(
    arr: Array,
    factor: int | None = None,
    probability: float | None = None,
) -> Array:
    """Decimate the data to a smaller number of rows.

    Must give either `factor` or `probability`.

    Parameters
    ----------
    arr : dask_awkward.Array
        Array collection to sample
    factor : int, optional
        if given, every Nth row will be kept. The counting restarts for each
        partition, so reducing the row count by an exact factor is not guaranteed
    probability : float, optional
        a number between 0 and 1, giving the chance of any particular
        row surviving. For instance, for probability=0.1, roughly 1-in-10
        rows will remain.

    """
    if (factor is None and probability is None) or (
        factor is not None and probability is not None
    ):
        raise ValueError("Give exactly one of factor or probability")
    if factor:
        return arr.map_partitions(lambda x: x[::factor], meta=arr._meta)
    assert probability is not None
    proba = float(probability)
    return arr.map_partitions(
        lambda x: x[_random_boolean_like(x, proba)], meta=arr._meta
    )
